<?php

namespace Drupal\content_model_documentation;

use Drupal\content_model_documentation\Entity\CMDocument;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ServiceProviderBase;
use Drupal\Core\Entity\EntityFieldManagerInterface;

/**
 * Service providing information on the Documentable entities.
 */
class DocumentableEntityProvider extends ServiceProviderBase {

  /**
   * Module configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Drupal's entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManager
   */
  protected $entityFieldManager;

  /**
   * An array of fieldmap.
   *
   * @var array
   */
  protected $fieldMap;

  /**
   * Constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configFactory
   *   Config factory service.
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entityFieldManager
   *   The time service.
   */
  public function __construct(ConfigFactoryInterface $configFactory, EntityFieldManagerInterface $entityFieldManager) {
    $this->config = $configFactory->get('content_model_documentation.settings');
    $this->entityFieldManager = $entityFieldManager;
    $this->fieldMap = $this->entityFieldManager->getFieldMap();
  }

  /**
   * Get all documentable entities (not including self).
   *
   * @return array
   *   Array of all entities identified as documentable.
   */
  public static function getUnDocumentedEntities(): array {
    $documentableEntityProvider = \Drupal::service('content_model_documentation.documentable.entity.provider');
    $documentable_entities = $documentableEntityProvider->getAvailableEntitiesToDocument();

    return $documentable_entities;
  }

  /**
   * Non-static run function to gather all the entities that are documentable.
   *
   * @return array
   *   An array of documentable entities.
   */
  public function getAvailableEntitiesToDocument() {
    $documentable_entities = CMDocument::getOtherDocumentableTypes();
    $documentable_entities = array_merge($documentable_entities, $this->getAllDocumentableEntities());
    $documentable_entities = $this->removeBaseEntities($documentable_entities);
    $documentable_entities = $this->removeDocumentedEntities($documentable_entities);

    return $documentable_entities;
  }

  /**
   * Generates an array of entity types if they are allowed to be documented.
   *
   * @return array
   *   An array whose elements in the form of 'entity name' => enabled status.
   */
  protected function getDocumentableEntityTypes(): array {
    $documentable_entities = [];
    $documentable_entities['block_content'] = $this->config->get('block');
    $documentable_entities['field'] = $this->config->get('field');
    $documentable_entities['media'] = $this->config->get('media');
    $documentable_entities['menu_link_content'] = $this->config->get('menu');
    $documentable_entities['node'] = $this->config->get('node');
    $documentable_entities['paragraph'] = $this->config->get('paragraph');
    $documentable_entities['taxonomy_term'] = $this->config->get('taxonomy');
    // Remove any that are disabled.
    $documentable_entities = array_filter($documentable_entities);
    $documentable_entities = array_keys($documentable_entities);

    return $documentable_entities;
  }

  /**
   * Gets all entities that are allowed by content_model_documentation config.
   *
   * @return array
   *   An array of key value pairs for the entities that are documentable.
   */
  protected function getAllDocumentableEntities(): array {
    $documentable_entities = [];
    $documentable_types = $this->getDocumentableEntityTypes();
    foreach ($documentable_types as $documentable_type) {
      $documentable_entities = $this->addEntityBundles($documentable_type, $documentable_entities);
      $documentable_entities = $this->addEntityFields($documentable_type, $documentable_entities);
    }
    // Lets sort them so the make more sense in the list.
    natcasesort($documentable_entities);
    return $documentable_entities;
  }

  /**
   * Removes base entities that make no sense to document and other cruft.
   *
   * @param array $documentable_entities
   *   An array of key value pairs to have base entities removed from.
   *
   * @return array
   *   An array of key value pairs for the entities that are documentable.
   */
  protected function removeBaseEntities(array $documentable_entities): array {
    $remove_these = [];
    // @todo Flesh this out So far there are none identified to remove.
    array_diff($documentable_entities, $remove_these);

    return $documentable_entities;
  }

  /**
   * Removes entities that have already been documented except self.
   *
   * @param array $documentable_entities
   *   An array of key value pairs to have base entities removed from.
   *
   * @return array
   *   An array of key value pairs for the entities that are documentable.
   */
  protected function removeDocumentedEntities(array $documentable_entities): array {
    $documented_entities = [];
    // @todo Flesh this out.  Needs to query existing and remove them.
    // Do not remove 'site.*' documents as those can have duplicates.
    // Can not remove itself, if this is an edit vs an add.
    array_diff($documentable_entities, $documented_entities);

    return $documentable_entities;
  }

  /**
   * Get documentable bundle entities and add them to $documentable_entities.
   *
   * @param string $entity_type
   *   The name of the entity type.
   * @param array $documentable_entities
   *   The incoming array of documentable entities.
   *
   * @return array
   *   Array of documentable entities.
   */
  protected function addEntityBundles($entity_type, array $documentable_entities): array {
    if (!empty($this->fieldMap[$entity_type]) && is_array($this->fieldMap[$entity_type])) {
      // We use uuid because everything has one.
      foreach ($this->fieldMap[$entity_type]['uuid']['bundles'] as $bundle) {
        $documentable_entities["{$entity_type}.{$bundle}"] = "{$entity_type}.{$bundle}";
      }
    }

    return $documentable_entities;
  }

  /**
   * Get documentable field entities and add them to $documentable_entities.
   *
   * @param string $entity_type
   *   The name of the entity type.
   * @param array $documentable_entities
   *   The incoming array of documentable entities.
   *
   * @return array
   *   Array of documentable entities.
   */
  protected function addEntityFields($entity_type, array $documentable_entities): array {
    $should_document_fields = $this->config->get('field');
    if ($should_document_fields && (!empty($this->fieldMap[$entity_type]))) {
      foreach ($this->fieldMap[$entity_type] as $element_name => $element) {
        if (!str_starts_with($element_name, 'field_')) {
          // It is not a field element, so bail out.
          continue;
        }
        foreach ($element['bundles'] as $bundle_name) {
          $documentable_entities["{$entity_type}.{$bundle_name}.{$element_name}"] = "{$entity_type}.{$bundle_name}.{$element_name}";
        }
      }

    }
    return $documentable_entities;
  }

}
